/*
 * Object for border traces using PamImage
 * Traces inside components, includes holes.
 * Input image must be strictly black/white.
 * Uses an intermediate 'side_table' containing between-pixel boundaries.
 * (C) Axel Brink, dept. of Artificial Intelligence, University of Groningen
 * October 2007.
 */

#include "bordertracinglib.h"
#include <iostream>
#include <cstring>

// *** PRIVATE ***

void BorderTracing::init()
{
    int image_type = image->getImageType();
    if ((image_type != BW_IMAGE) && (image_type != GRAY_IMAGE)) {
        std::cerr << "Error: " << image_type << " is an illegal image type. BW_IMAGE ("
                  << BW_IMAGE << ") or GRAY_IMAGE (" << GRAY_IMAGE << ") is required." << std::endl;
        throw 1;
    }
    
    width = image->getWidth();
    height = image->getHeight();
    
    // Initialize 1D array for trace data.
    // Note that every pixel could be possibly used as inside and outside trace.
    trace_data = new Pos[2 * width * height];
    free_data_ptr = trace_data;
    
    // Initialize temporary 2D array for between-pixel borders.
    // An extra row, column is needed for bottom, right border info.
    side_table = new PixelSide* [height + 1];
    for (int y = 0; y <= height; ++y) {
        side_table[y] = new PixelSide[width + 1];
        for (int x = 0; x <= width; ++x) {
            side_table[y][x].left = false;
            side_table[y][x].top = false;
        }
    }
}

// compute where white/black and black/white transitions occur
void BorderTracing::compute_side_table()
{
    GrayPixel** pixels = image->getGrayPixels();
    // *** top sides ***
    // top and bottom image border (are boundary where adjacent to black pixel)
    for (int x = 0; x < width; ++x) {
        if (pixels[0][x]          == 0) side_table[0][x].top = true;
        if (pixels[height - 1][x] == 0) side_table[height][x].top = true;
    }
    // rest of the image
    for (int y = 1; y < height; ++y) {
        for (int x = 0; x < width; ++x) {
            if (pixels[y][x] != pixels[y - 1][x]) side_table[y][x].top = true;
        }
    }
    
    // *** left sides ***
    // left and right image border (are boundary where adjacent to black pixel)
    for (int y = 0; y < height; ++y) {
        if (pixels[y][0]       == 0) side_table[y][0].left = true;
        if (pixels[y][width-1] == 0) side_table[y][width].left = true;
    }
    // rest of the image
    for (int y = 0; y < height; ++y) {
        for (int x = 1; x < width; ++x) {
            if (pixels[y][x] != pixels[y][x - 1]) side_table[y][x].left = true;
        }
    }
}

inline void add_pos(BorderTrace &trace, int x, int y) {
    Pos pos;
    pos.x = x;
    pos.y = y;
    trace.list[trace.len] = pos;
    trace.len++;
}

// Add trace starting here (a side is above)
void BorderTracing::trace_here(int x, int y) {
    // PRE: a horizontal side is above pixel (y, x)
    //      and side_table contains all sides
    BorderTrace trace;
    trace.len = 0;
    trace.list = free_data_ptr;
    GrayPixel** pixels = image->getGrayPixels();
    int old_dir;
    
    // Set dir and x such that the known side will be 'discovered' first.
    if (y == height) {
        // utter bottom; object is above the side
        old_dir = 1;
        side_table[y-1][x].left = false; // avoid revisiting the same pixel at the end
    } else if (pixels[y][x] != 0) {
        // object is above the side
        old_dir = 1;
        side_table[y-1][x].left = false; // avoid revisiting the same pixel at the end
    } else {
        // object is below the side
        old_dir = 3;
        side_table[y][x+1].left = false; // avoid revisiting the same pixel at the end
        x++;
    }
    // upper-left corner of (x, y) is starting point of boundary and dir is correct
    
    bool side_found;
    do {
        side_found = false;
        // find a new side starting from upper left corner of pixel (x, y)
        int dir = old_dir;
        for (int i = 0; ((i < 4) && (!side_found)); ++i) {
            dir = (dir + 3) % 4;
            switch (dir) {
                case 0: // right
                    if (x < width) {
                        if (side_table[y][x].top) {
                            side_table[y][x].top = false;
                            side_found = true;
                            if (old_dir != 3) add_pos (trace, x, y - 1);
                            x++;
                        }
                    }
                    break;
                case 1: // up
                    if (y > 0) {
                        if (side_table[y-1][x].left) {
                            side_table[y-1][x].left = false;
                            side_found = true;
                            if (old_dir != 0) add_pos (trace, x - 1, y - 1);
                            y--;
                        }
                    }
                    break;
                case 2: // left
                    if (x > 0) {
                        if (side_table[y][x-1].top) {
                            side_table[y][x-1].top = false;
                            side_found = true;
                            if (old_dir != 1) add_pos (trace, x - 1, y);
                            x--;
                        }
                    }
                    break;
                case 3: // down
                    if (y < height) {
                        if (side_table[y][x].left) {
                            side_table[y][x].left = false;
                            side_found = true;
                            if (old_dir != 2) add_pos (trace, x, y);
                            y++;
                        }
                    }
                    break;
            }
        }
        old_dir = dir;
    } while (side_found);
    
    traces.push_back(trace);
    free_data_ptr = trace.list + trace.len; // free place for next trace
}
 
void BorderTracing::trace_borders()
{
    compute_side_table();
    
    // search for new starting point (a horizontal side)
    for (int y = 0; y <= height; ++y) {
        for (int x = 0; x <= width; ++x) {
            if (side_table[y][x].top) trace_here(x, y);
        }
    }
    // delete[] side_table; --> could be done here already! (kept for drawing)
}

// *** PUBLIC ***

BorderTracing::BorderTracing(PamImage* im) {
    //image = im->convert(BW_IMAGE); // creates a new image in memory
    image = im;
    init();
    trace_borders();
}

BorderTracing::~BorderTracing() {
    delete[] trace_data;
    for (int y = 0; y <= height; ++y) {
        delete[] side_table[y];
    }
    delete[] side_table;
    //delete image;
}

// Return a vector of all border traces
std::vector<BorderTrace> BorderTracing::getTraces() {
    return traces;
}

PamImage* BorderTracing::draw() {
    PamImage* out_im = image->convert(RGB_IMAGE);
    RGBPixel** pixels = out_im->getRGBPixels();
    
    // commented text is for movie creation
    //int imc = 0;
    //char filename[20]; 

    for(unsigned int t = 0; t < traces.size(); ++t) {
        //std::cerr << " trace " << t << " of " << traces.size() << "..." << std::endl;
        BorderTrace trace = traces[t];
        Pos pos = trace.list[0];
        pixels[pos.y][pos.x].r = 255;
        for (int p = 1; p < trace.len; ++p) {
            Pos pos = trace.list[p];
            pixels[pos.y][pos.x].g = 255;
            
            //sprintf(filename, "/tmp/movie/im%04d.ppm", imc++);
            //if (imc < 10000) out_im->save(filename);
        }
    }
    return out_im;
}

PamImage* BorderTracing::draw_sides() {
    PamImage* out_im = new PamImage(RGB_IMAGE, width + 1, height + 1);
    RGBPixel** pixels = out_im->getRGBPixels();

    for (int y = 0; y <= height; ++y) {
        for (int x = 0; x <= width; ++x) {
            if (side_table[y][x].left) pixels[y][x].r = 255;
            if (side_table[y][x].top) pixels[y][x].g = 255;
        }
    }
    
    return out_im;
}
