/* This source file is a part of 
 * 
 * Document feature library (DFLib)
 * 
 * @copyright: Sheng He, University of Groningen (RUG)
 * 
 * Email: heshengxgd@gmail.com
 * 
 *  27, Nov., 2015
 * 
 * 
 *    Context Of Lines Discriptor for handwritten documents
 * */

#include "dflCOLDfeatlib.h" 
#include <cmath>

//#define DEBUG 1

#ifdef DEBUG
#include "pamImage.h"
#endif

COLDfeat::COLDfeat(float _scale,int _nrad,int _nang)
{
	scale = _scale;
	nrad = _nrad;
	nang = _nang;
	
	silence = false;
}

COLDfeat::~COLDfeat()
{
}

std::vector<linevec> COLDfeat::getLineVector(std::vector<std::vector<dflPixel> >& keypointlist,int kLength)
{
	std::vector<linevec> veclist;
	
	int dx,dy;
	
	int nl = kLength;
	//for(int nl = 1; nl <= kLength; nl++)
	{
		for(int np = 0; np < (int)keypointlist.size(); ++np)
		{
			int sizep = keypointlist[np].size();
			if( sizep < nl+1) 
				continue;
			
			std::vector<dflPixel> curlist = keypointlist[np];
			
			for(int k = 0; k < (int)curlist.size(); ++k)
			{
				int next = k + nl;
				if(next >= (int)curlist.size())
					next -= curlist.size();
				
				dx = curlist[next].x - curlist[k].x;
				dy = curlist[next].y - curlist[k].y;
				
				bool found = false;
				int findx = -1;
				for(int sv = 0; sv < (int)veclist.size(); ++sv)
				{
					if(veclist[sv].point.y == dy && veclist[sv].point.x == dx)
					{
						found = true;
						findx = sv;
						break;
					}
				}
				
				if(found)
					veclist[findx].number++;
				else
				{
					linevec lv;
					lv.point.x = dx;
					lv.point.y = dy;
					lv.number = 1;
					veclist.push_back(lv);
				}
			}	
		}
	}
		
	return veclist;
}

/** model == 1 compute the angle histogram
 *  model == 2 compute the line histogram
 *  model == 3 compute the concatenate of angle and line histogram
 * */
std::vector<float> COLDfeat::getCOLDLenAngleSeperately(std::vector<linevec>& veclist,int model)
{
	double lpbase = pow(10,log10((double)nrad)/nrad);
	double angbase = 2 * M_PI / (double) nang;
			
	std::vector<double> radiiQuants(nrad);	

	int nloss = 0;
		
	for (int i=0; i<nrad; ++i){
		radiiQuants[i] = (pow(lpbase,i+1)-1)/(nrad - 1);
	}
	
	double s_radius = (double)scale / radiiQuants[0];

	for(int i = 0; i < nrad; ++i){	
		radiiQuants[i] = radiiQuants[i] * s_radius + scale;
	}
	
	std::vector<float> featline(nrad,0.0);
	std::vector<float> featangle(nang,0.0);
	
	for(int nv = 0; nv < (int)veclist.size(); ++nv)
	{
		float ang = atan2(veclist[nv].point.y,veclist[nv].point.x);
		if(ang < 0.0) ang += M_PI*2;
		
		float e_radius = sqrt(veclist[nv].point.y*veclist[nv].point.y+veclist[nv].point.x*veclist[nv].point.x);
		
		if(e_radius > radiiQuants[nrad-1]) {
			nloss++;
			continue;
		}
		
		int angbin = ang / angbase;
		if(angbin >= nang) angbin--;
		if(angbin < 0) angbin = 0;
		
		int radbin = -1;
		for(int r = 0; r < nrad; ++r)
		{
			if(e_radius < radiiQuants[r])
			{
				radbin = r;
				break;
			}
		}
		
		if(angbin >=0 && angbin < nang)
			featangle[angbin] += veclist[nv].number;
		
		if(radbin >= 0 && radbin < nrad)
			featline[radbin] += veclist[nv].number;
	}
	
	if(silence)std::cout<<"loss points: "<<(float)nloss/(float)veclist.size()*100<<"\%\n";
	
	std::vector<float> featvec;
	double sum = 0; 
	for(int n = 0; n < featangle.size(); ++n)
		sum += featangle[n];
	for(int n = 0; n < featangle.size(); ++n){
		featangle[n] /= sum;
		if(model==1 || model==3) 
			featvec.push_back(featangle[n]);
	}
	
	sum = 0.0;
	for(int n = 0; n < featline.size(); ++n)
		sum += featline[n];
	for(int n = 0; n < featline.size(); ++n){
		featline[n] /= sum;
		if(model==2 || model==3) 
			featvec.push_back(featline[n]);
	}
	
	return featvec;
}


std::vector<float> COLDfeat::getCOLDLogPolar(std::vector<linevec>& veclist)
{
	double lpbase = pow(10,log10((double)nrad)/nrad);
	double angbase = 2 * M_PI / (double) nang;
			
	std::vector<double> radiiQuants(nrad);	

	std::vector<std::vector<float> > featvec(nrad);
	for(int n = 0; n < nrad; ++n)
		featvec[n].resize(nang,0.0);
	
	int nloss = 0;
		
	for (int i=0; i<nrad; ++i){
		radiiQuants[i] = (pow(lpbase,i+1)-1)/(nrad - 1);
	}
	
	double s_radius = (double)scale / radiiQuants[0];

	for(int i = 0; i < nrad; ++i){	
		radiiQuants[i] = radiiQuants[i] * s_radius + scale;
	}

#ifdef DEBUG
	if(silence) std::cout<<"Starting to debug!\n";
	
	for(int i = 0; i < nrad; ++i)	
		std::cout<<i+1<<"-th radius: "<<radiiQuants[i]/radiiQuants[nrad-1]*10.5<<"\n";
		
	int center = radiiQuants[nrad-1]+0.5;
	const int w = 2*center+1;
	const int h = 2*center+1;
	PamImage* im = new PamImage(3,w,h);
	RGBPixel** impt = im->getRGBPixels();
	for(int my = 0; my < h; ++my)
		for(int mx = 0; mx < w; ++mx){
			impt[my][mx].r=impt[my][mx].b=impt[my][mx].g=255;
			//impt[my][mx].r=0;
			//impt[my][mx].g=0;
			//impt[my][mx].b=255;
		}
	
	double max_num = 0;
	for(int nv = 0; nv < veclist.size(); ++nv)
	{
		float e_radius = sqrt(veclist[nv].point.y*veclist[nv].point.y+veclist[nv].point.x*veclist[nv].point.x);
		
		if(e_radius > radiiQuants[nrad-1]) {
			continue;
		}
		else
		{
			 if(veclist[nv].number>max_num)
				max_num = veclist[nv].number;
		}
	}
	
	int rdis = 2 * radiiQuants[nrad-1] + 0.5;
	
	int cpoint = rdis / 2;
	
	std::vector<std::vector<float> > data(rdis);
	for(int n = 0; n < rdis; ++n)
		data[n].resize(rdis,0.0);
	
	for(int nv = 0; nv < veclist.size(); ++nv)
	{
		float e_radius = sqrt(veclist[nv].point.y*veclist[nv].point.y+veclist[nv].point.x*veclist[nv].point.x);
		
		if(e_radius > radiiQuants[nrad-1]) {
			continue;
		}
		else
			data[veclist[nv].point.y+cpoint][veclist[nv].point.x+cpoint] = veclist[nv].number;
	}
	
	
	/* smooth */
	int kernel_size = 15;
	int kcenter = (kernel_size-1)/2;
	float sigma = 5.0;
	
	std::vector<std::vector<float> > kernel(kernel_size);
	for(int nk = 0; nk < kernel_size; ++nk)
		kernel[nk].resize(kernel_size,0.0);
	float k_sum = 0;
	for(int ny = 0; ny < kernel_size; ++ny)
		for(int nx = 0; nx < kernel_size; ++nx){
			kernel[ny][nx] = exp(- (float)((ny-kcenter)*(ny-kcenter)+(nx-kcenter)*(nx-kcenter))/sigma);
			k_sum += kernel[ny][nx];
	}
	
	for(int ny = 0; ny < kernel_size; ++ny)
		for(int nx = 0; nx < kernel_size; ++nx)
			kernel[ny][nx] /= k_sum;
	
	
	std::vector<std::vector<float> > sdata(rdis);
	for(int n = 0; n < rdis; ++n)
		sdata[n].resize(rdis,0.0);
				
	for(int ry = 0; ry < rdis; ++ry)
		for(int rx = 0; rx < rdis; ++rx)
	{
		float value = 0.0;
		float c_sum = 0.0;
		for(int ny = -kcenter; ny <= kcenter; ++ny)
			for(int nx = -kcenter; nx <=kcenter; ++nx)
		{
			if(ry+ny < 0 || ry+ny >= rdis || rx+nx < 0 || rx+nx >= rdis)
				continue;
			value += kernel[ny+kcenter][nx+kcenter]*data[ry][rx];
			c_sum += kernel[ny+kcenter][nx+kcenter];
		}
		
		value /= c_sum;
		sdata[ry][rx] = value;
	}
	
	float max_data = -HUGE;
	for(int ry = 0; ry < rdis; ++ry)
		for(int rx = 0; rx < rdis; ++rx)
			if(sdata[ry][rx]>max_data)
				max_data = sdata[ry][rx];
				
	for(int ry = 0; ry < rdis; ++ry)
		for(int rx = 0; rx < rdis; ++rx)
	{
		float value = sdata[ry][rx]/max_data*255;
		if(value < 10) 
		{
			impt[ry][rx].r=impt[ry][rx].g=impt[ry][rx].b=255;
		}
		else
		{
			impt[ry][rx].r = (int)value;
			impt[ry][rx].g = 0;
			impt[ry][rx].b = 255-(int)value;
		}
	}
	/*for(int nv = 0; nv < veclist.size(); ++nv)
	{
		float e_radius = sqrt(veclist[nv].point.y*veclist[nv].point.y+veclist[nv].point.x*veclist[nv].point.x);
		
		if(e_radius > radiiQuants[nrad-1]) {
			continue;
		}
		else
		{
			int color = (int)((double)veclist[nv].number/max_num*255+60);
			if(color>255)color=255;
			
			impt[veclist[nv].point.y+center][veclist[nv].point.x+center].r = color;
			impt[veclist[nv].point.y+center][veclist[nv].point.x+center].g = 0;//255-(int)((double)veclist[nv].number/max_num*255);
			impt[veclist[nv].point.y+center][veclist[nv].point.x+center].b = 255-color;//255-(int)((double)veclist[nv].number/max_num*255);
		}
	}
	
	/*for(int my = 0; my < h; ++my)
		for(int mx = 0; mx < w; ++mx)
	{
		int cy = my - center;
		int cx = mx - center;
		float dis = sqrt(cy*cy+cx*cx);
		bool found = false;
		for(int nr = 0; nr < nrad; ++nr)
		{
			if(std::fabs(dis-radiiQuants[nr])<0.5)
				found = true;
		}
		if(found) {
			impt[my][mx].r = 0;
			impt[my][mx].g = 255;
			impt[my][mx].b = 0;
		}
	}*/
	
	im->save("debug_cold.ppm");
	delete im;
	
#endif	
	for(int nv = 0; nv < (int)veclist.size(); ++nv)
	{
		float ang = atan2(veclist[nv].point.y,veclist[nv].point.x);
		if(ang < 0.0) ang += M_PI*2;
		
		float e_radius = sqrt(veclist[nv].point.y*veclist[nv].point.y+veclist[nv].point.x*veclist[nv].point.x);
		
		if(e_radius > radiiQuants[nrad-1]) {
			nloss++;
			continue;
		}
		
		int angbin = ang / angbase;
		if(angbin >= nang) angbin--;
		if(angbin < 0) angbin = 0;
		
		int radbin = -1;
		for(int r = 0; r < nrad; ++r)
		{
			if(e_radius < radiiQuants[r])
			{
				radbin = r;
				break;
			}
		}
		
		if(radbin >= 0 && radbin < nrad)
			featvec[radbin][angbin] += veclist[nv].number;
	}
	
	if(silence)std::cout<<"loss points: "<<(float)nloss/(float)veclist.size()*100<<"\%\n";
	
	std::vector<float> feature;
	double sum = 0.0;
	for(int nr = 0; nr < nrad; ++nr)
		for(int nv = 0; nv < nang; ++nv)
			sum += featvec[nr][nv];
	if(sum==0) sum = 1;
	for(int nr = 0; nr < nrad; ++nr){
		for(int nv = 0; nv < nang; ++nv){
			feature.push_back(featvec[nr][nv]/sum);
#ifdef DEBUG
			std::cout<<nv<<"/"<<featvec[nr][nv]/sum*1000<<", ";
#endif
		}
#ifdef DEBUG
		std::cout<<"\n";
#endif
	}
	return feature;
}

std::vector<float> COLDfeat::getCOLDfeat(std::vector<std::vector<dflPixel> >& keypointlist,int kLength)
{
	
	std::vector<float> feature;
	
	/*if(true)
	{
		std::vector<linevec> veclist = getLineVector(keypointlist,kLength);
		std::vector<float> feat = getCOLDLogPolar(veclist);
		return feat;
	}*/
	
	for(int nk = 1; nk <= kLength; ++nk)
	{
		std::vector<linevec> veclist = getLineVector(keypointlist,nk);
		std::vector<float> feat = getCOLDLogPolar(veclist);
		for(int nf = 0; nf < (int)feat.size(); ++nf)
			feature.push_back(feat[nf]);
	}
	
	return feature;
}

std::vector<float> COLDfeat::getCOLDfeat(PamImage* binary, int kLength,float threshold)
{
	BorderTracing tr(binary);
	std::vector<BorderTrace> traces = tr.getTraces();
	std::vector<std::vector<dflPixel> > kplist;
		
	int minimum_len = 15;
		
	for(int n = 0; n < (int)traces.size(); ++n)
	{
		
		if(traces[n].len < minimum_len) continue;
		
		std::vector<dflPixel> contour;
		for(int p = 0; p < traces[n].len; ++p)
		{
			dflPixel pp;
			pp.x = traces[n].list[p].x;
			pp.y = traces[n].list[p].y;
			contour.push_back(pp);
		}
		
		dflContourlinefit linfit(0,true,threshold);
		std::vector<dflPixel> keypoints = linfit.linefit(contour);
		kplist.push_back(keypoints);
	}
	
	return getCOLDfeat(kplist,kLength);
}

std::vector<float> COLDfeat::getCOLDfeatLineAngleSeperately(PamImage* binary, int kLength,float threshold,int model)
{
	BorderTracing tr(binary);
	std::vector<BorderTrace> traces = tr.getTraces();
	std::vector<std::vector<dflPixel> > kplist;
		
	int minimum_len = 15;
		
	for(int n = 0; n < (int)traces.size(); ++n)
	{
		
		if(traces[n].len < minimum_len) continue;
		
		std::vector<dflPixel> contour;
		for(int p = 0; p < traces[n].len; ++p)
		{
			dflPixel pp;
			pp.x = traces[n].list[p].x;
			pp.y = traces[n].list[p].y;
			contour.push_back(pp);
		}
		
		dflContourlinefit linfit(0,true,threshold);
		std::vector<dflPixel> keypoints = linfit.linefit(contour);
		kplist.push_back(keypoints);
	}
	
	std::vector<float> feature;
	

	for(int nk = 1; nk <= kLength; ++nk)
	{
		std::vector<linevec> veclist = getLineVector(kplist,nk);
		std::vector<float> feat = getCOLDLenAngleSeperately(veclist,model);
		for(int nf = 0; nf < (int)feat.size(); ++nf)
			feature.push_back(feat[nf]);
	}
	
	return feature;
	
}
