/* 
 * This code is the C++ version of the method published on:
 * 
 * Dilip K. Prasad, Chai Quek, Maylor K.H. Leung, and Siu-Yeung Cho, “A parameter independent line fitting method,” 1st IAPR Asian Conference on Pattern Recognition 
 * 
 * Please cite this paper if you use this program
 * 
 ***  This source file is a part of *********
 * 
 * Document feature library (DFLib)
 * 
 * @copyright: Sheng He, University of Groningen (RUG)
 * 
 * Email: heshengxgd@gmail.com
 * 
 *  27, Nov., 2015
 * */


#include "dflContourlinefitlib.h"
#include <cmath>
#include <iostream>


dflContourlinefit::dflContourlinefit(int in_verbose, bool in_removeLinear,float in_collinearThresh)
{
	verbose=in_verbose;
	romoveCollinear = in_removeLinear;
	collinearThresh = in_collinearThresh;
	if(verbose == 0)
		silence = true;
	else
		silence = false;		
}

dflContourlinefit::~dflContourlinefit()
{
}

std::vector<dflPixel> dflContourlinefit::linefit(std::vector<dflPixel>& contour)
{
	int first = 0;
	int last = contour.size()-1;
	
	std::vector<dflPixel> keypoints;
	
	double max_dev;
	int index;
	double D_temp;
	double precision;
	double reliability;
	double del_tol_max;
	double sum_dev;
	
	if(verbose>0)
		std::cout<<"Starting to compute the contour dominant points!\n";
	//keypoints.push_back(contour[0]);
		
	while(first < last)
	{
		std::vector<dflPixel> subcontour;
		for(int n = first; n <= last; ++n)
			subcontour.push_back(contour[n]);
		
		//if(verbose>1)std::cout<<"fist:("<<first<<" end:"<<last<<")\n";
		maxlinedev_opt(subcontour,&max_dev,&index,&D_temp,&precision,&reliability,&del_tol_max,&sum_dev);
		
		while(max_dev > del_tol_max)
		{
			last=index+first;
			//if(verbose>1)std::cout<<"fist:("<<first<<" end:"<<last<<")\n";
			std::vector<dflPixel> subcontour;
			if(last<=first)break;
			for(int n = first; n <= last; ++n)
				subcontour.push_back(contour[n]);
			
			
			maxlinedev_opt(subcontour,&max_dev,&index,&D_temp,&precision,&reliability,&del_tol_max,&sum_dev);
			
		}
		
		if(verbose>1)
			std::cout<<"Find keypoint: ("<<contour[last].y<<" "<<contour[last].x<<")\n";
			
		keypoints.push_back(contour[last]);
		first=last;
		last=contour.size()-1;
	}
	
	
	if(romoveCollinear) removeCollinearPoints(keypoints);
	return keypoints;
}

void dflContourlinefit::maxlinedev_opt(std::vector<dflPixel>& subcontour,double *max_dev,int *index,double *D_temp,double *precision,double *reliability,double *del_tol_max,double *sum_dev)
{
	int first = 0;
	int last = subcontour.size()-1;
	double a1 = double(subcontour[first].y-subcontour[last].y)/double(subcontour[first].y*subcontour[last].x-subcontour[last].y*subcontour[first].x);
	double a2 = double(subcontour[first].x-subcontour[last].x)/double(subcontour[first].x*subcontour[last].y-subcontour[last].x*subcontour[first].y);
	

	std::vector<double> dev;
	
	if( std::isnan(a1) && std::isnan(a2) )
	{
		for(int n = 0; n < subcontour.size(); ++n)
		{
			double dx = subcontour[n].x-subcontour[0].x;
			double dy = subcontour[n].y-subcontour[0].y;
			double value = std::fabs(std::sqrt(dx*dx+dy*dy));
			dev.push_back(value);
		}
	}
	else if(std::isinf(a1) && std::isinf(a2) )
	{
		double c = (double)subcontour[0].x/(double)subcontour[0].y;
		double sc = std::sqrt(1.0+c*c);
		for(int n = 0; n < subcontour.size(); ++n)
		{
			double value1 = (double)subcontour[n].x/sc;
			double value2 = -c*(double)subcontour[n].y/sc;
			double value = std::fabs(value1+value2);
			dev.push_back(value);
		}
	}
	else
	{
		double div = std::sqrt(a1*a1+a2*a2);
		for(int n = 0; n < subcontour.size(); ++n)
		{
			double value = (double)subcontour[n].x*a1+(double)subcontour[n].y*a2-1;
			value /= div;
			dev.push_back(std::fabs(value));
		}
	}
	
	double d_max = -HUGE;
	int index_d_max_v = -1;
	
	for(int n = 0; n < dev.size(); ++n)
	{
		if(dev[n]>d_max)
		{
			d_max = dev[n];
			index_d_max_v = n;
		}
	}
	
	if(index_d_max_v<0)
	{
		if(!silence)std::cout<<"Warning: there is no minimum value!\n";
	}
	
	double precision_v = 0.0;
	double norm_v = 0.0;
	double sum_dev_v = 0.0;
	for(int n = 0; n < dev.size(); ++n){
		norm_v += dev[n]*dev[n];
		sum_dev_v += dev[n];
	}
	
	norm_v = std::sqrt(norm_v);
	precision_v = norm_v/std::sqrt(double(last+1));
	
	double S_max_v = -HUGE;
	for(int n = 0; n < subcontour.size(); ++n)
	{
		double dx = subcontour[n].x-subcontour[0].x;
		double dy = subcontour[n].y-subcontour[0].y;
		double value = std::sqrt(dx*dx+dy*dy);
		if(value > S_max_v)
			S_max_v = value;
	}
	
	
	double reliability_v = sum_dev_v/S_max_v;
	//std::cout<<"input:"<<S_max_v<<" ";
	double del_phi_max_v = function_digital_case_with_max_error(S_max_v);
	//std::cout<<"output:"<<del_phi_max_v<<" ";
	double del_tol_max_v = std::tan(del_phi_max_v)*S_max_v;
	//std::cout<<"Final: "<<del_tol_max_v<<"\n";
	*max_dev = d_max;
	*index = index_d_max_v;
	*D_temp = S_max_v;
	*precision = precision_v;
	*reliability = reliability_v;
	*del_tol_max = del_tol_max_v;
	*sum_dev = sum_dev_v;
}

void dflContourlinefit::removeCollinearPoints(std::vector<dflPixel>& points)
{
	
	if(collinearThresh<=0) 
	{
		if(!silence)std::cout<<"The threshold: "<<collinearThresh<<" of removing collinear points are incorrect!\n";
		return;
	}
	
	float disx,disy;
	
	while(true)
	{
		int sc = points.size();
		if(sc < 3) break;
		
		std::vector<float> dislist(sc,0.0);
		for(int idx = 0; idx < points.size(); ++idx)
		{
			dflPixel curr,prev,next;
			if(idx >= sc) idx = 0;
			curr = points[idx];
			prev = points[(idx-1+sc)%sc];
			next = points[(idx+1)%sc];
			disx = next.x-prev.x;
			disy = next.y-prev.y;
			dislist[idx] = std::fabs(disx*(prev.y-curr.y)-disy*(prev.x-curr.x))/sqrt(disx*disx+disy*disy);
		}
		double min = HUGE;
		int idmin = -1;
		for(int idn = 0; idn < dislist.size(); ++idn)
		{
			if(dislist[idn]<min)
			{
				min = dislist[idn];
				idmin = idn;
			}
		}
		
		//std::cout<<"min: "<<min<<"\n";
		if(min > collinearThresh)
			break;
		
		if(idmin < 0 && idmin >= sc)
		{
			if(!silence)std::cout<<"Warning: can not find the minimum value!\n";
		}
		else{
			points.erase(points.begin()+idmin);
			if(verbose>1)std::cout<<"one point has been removed!\n";
		}
	}
}

double dflContourlinefit::function_digital_case_with_max_error(double S_max)
{
	std::vector<double> phii;
	double piStep = M_PI/360.0;
	
	for(int n = 0; n < 721; ++n)
		phii.push_back(n*piStep);
	
	int psz = phii.size();
	std::vector<double> term11(psz,0.0);
	std::vector<double> term12(psz,0.0);
	std::vector<double> term13(psz,0.0);
	std::vector<double> term14(psz,0.0);
	std::vector<double> term15(psz,0.0);
	std::vector<double> term16(psz,0.0);
	std::vector<double> term17(psz,0.0);
	std::vector<double> term18(psz,0.0);
	
	for(int n = 0; n < psz; ++n)
	{
		term11[n] = std::fabs(std::cos(phii[n]));
		term12[n] = std::fabs(std::sin(phii[n]));
		term13[n] = std::fabs(std::sin(phii[n])+std::cos(phii[n]));
		term14[n] = std::fabs(std::sin(phii[n])-std::cos(phii[n]));
		
		term15[n] = std::fabs(std::cos(phii[n]));
		term16[n] = std::fabs(std::sin(phii[n]));
		term17[n] = std::fabs(std::sin(phii[n])+std::cos(phii[n]));
		term18[n] = std::fabs(std::sin(phii[n])-std::cos(phii[n]));
	}
	
	std::vector<double> tt21(psz,0.0);
	std::vector<double> tt22(psz,0.0);
	std::vector<double> tt23(psz,0.0);
	std::vector<double> tt24(psz,0.0);
	std::vector<double> tt25(psz,0.0);
	std::vector<double> tt26(psz,0.0);
	std::vector<double> tt27(psz,0.0);
	std::vector<double> tt28(psz,0.0);
	
	for(int n = 0; n < psz; ++n)
	{
		tt21[n] = std::sin(phii[n])/S_max;
		tt22[n] = std::cos(phii[n])/S_max;
		tt23[n] = (std::sin(phii[n])-std::cos(phii[n]))/S_max;
		tt24[n] = (std::sin(phii[n])+std::cos(phii[n]))/S_max;
		
		tt25[n] = -std::sin(phii[n])/S_max;
		tt26[n] = -std::cos(phii[n])/S_max;
		tt27[n] = -(std::sin(phii[n])-std::cos(phii[n]))/S_max;
		tt28[n] = -(std::sin(phii[n])+std::cos(phii[n]))/S_max;
		
	}
	
	std::vector<double> term21(psz,0.0);
	std::vector<double> term22(psz,0.0);
	std::vector<double> term23(psz,0.0);
	std::vector<double> term24(psz,0.0);
	std::vector<double> term25(psz,0.0);
	std::vector<double> term26(psz,0.0);
	std::vector<double> term27(psz,0.0);
	std::vector<double> term28(psz,0.0);
	
	for(int n = 0; n < psz; ++n)
	{
		term21[n] = S_max*(1-tt21[n]+tt21[n]*tt21[n]);
		term22[n] = S_max*(1-tt22[n]+tt22[n]*tt22[n]);
		term23[n] = S_max*(1-tt23[n]+tt23[n]*tt23[n]);
		term24[n] = S_max*(1-tt24[n]+tt24[n]*tt24[n]);
		term25[n] = S_max*(1-tt25[n]+tt25[n]*tt25[n]);
		term26[n] = S_max*(1-tt26[n]+tt26[n]*tt26[n]);
		term27[n] = S_max*(1-tt27[n]+tt27[n]*tt27[n]);
		term28[n] = S_max*(1-tt28[n]+tt28[n]*tt28[n]);
	}

	std::vector<double> case_value1(psz,0.0);
	std::vector<double> case_value2(psz,0.0);
	std::vector<double> case_value3(psz,0.0);
	std::vector<double> case_value4(psz,0.0);
	std::vector<double> case_value5(psz,0.0);
	std::vector<double> case_value6(psz,0.0);
	std::vector<double> case_value7(psz,0.0);
	std::vector<double> case_value8(psz,0.0);
	
	for(int n = 0; n < psz; ++n)
	{
		case_value1[n] = 1.0/(S_max*S_max)*term11[n]*term21[n];
		case_value2[n] = 1.0/(S_max*S_max)*term12[n]*term22[n];
		case_value3[n] = 1.0/(S_max*S_max)*term13[n]*term23[n];
		case_value4[n] = 1.0/(S_max*S_max)*term14[n]*term24[n];
		case_value5[n] = 1.0/(S_max*S_max)*term15[n]*term25[n];
		case_value6[n] = 1.0/(S_max*S_max)*term16[n]*term26[n];
		case_value7[n] = 1.0/(S_max*S_max)*term17[n]*term27[n];
		case_value8[n] = 1.0/(S_max*S_max)*term18[n]*term28[n];
	}
	
	std::vector<double> del_phi_tilde_max(psz,0.0);
	for(int n = 0; n < psz; ++n)
	{
		double max = -HUGE;
		if(case_value1[n]>max)max=case_value1[n];
		if(case_value2[n]>max)max=case_value2[n];
		if(case_value3[n]>max)max=case_value3[n];
		if(case_value4[n]>max)max=case_value4[n];
		if(case_value5[n]>max)max=case_value5[n];
		if(case_value6[n]>max)max=case_value6[n];
		if(case_value7[n]>max)max=case_value7[n];
		if(case_value8[n]>max)max=case_value8[n];
		del_phi_tilde_max[n]=max;
	}
	
	double max_res = -HUGE;
	
	for(int n = 0; n < psz; ++n)
		if(del_phi_tilde_max[n]>max_res)
			max_res=del_phi_tilde_max[n];
			
	return max_res;
}
